#ifndef __KERNEL__
#define __KERNEL__
#endif

#ifndef MODULE
#define MODULE
#endif

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>

#include <linux/types.h>
#include <linux/string.h>

#include <linux/netfilter_ipv4.h>
#include <linux/timer.h>
#include <linux/sched.h>

#include <asm/atomic.h>
#include <asm/uaccess.h>

#include "detector.h"

extern int recovery(unsigned int flag, struct task_struct * task, int cpu, const char* msg);

static struct det_msg dm;

static struct task_struct* max_exec_task;
static u64 max_exec; 
static struct task_struct* max_mem_task;
static unsigned long max_mem;

static inline int is_kthread(struct task_struct* task) {
  return (task->mm == NULL);
}

static int check_exec(struct task_struct* task) {
  /* ignore sleeping process and kernel thread */
  if (task->state != TASK_RUNNING || is_kthread(task))
    return 0;
  u64 exec_max = task->se.statistics.exec_max;
  if (exec_max > max_exec) {
    max_exec = exec_max;
    max_exec_task = task;
  }
  return (max_exec > EXEC_TIMEOUT);
}

static inline unsigned long get_mm_rss_counter(struct mm_struct* mm, int member) {
  return atomic_long_read(&mm->rss_stat.count[member]);
}

static int check_mem(struct task_struct* task) {
  if (is_kthread(task))
    return 0;

  struct mm_struct* mm = get_task_mm(task);
  /*
  unsigned long kb_res, kb_swap;
  kb_res = (get_mm_rss_counter(mm, MM_FILEPAGES) + 
      get_mm_rss_counter(mm, MM_ANONPAGES)) << 
      (PAGE_SHIFT-10);
  kb_swap = (get_mm_rss_counter(mm, MM_SWAPENTS)) << (PAGE_SHIFT-10);

  if (kb_res + kb_swap > max_mem) {
    max_mem = kb_res + kb_swap;
    max_mem_task = task;
  }
  */
  if (mm->total_vm > max_mem) {  /* should be improved, not precise enough */
    max_mem = mm->total_vm;
    max_mem_task = task;
  }
  return (max_mem > MAX_MEM_OCPY);
}

/**
 * Poll all task and find the most suspicious task
 */
static int poll_suspicious_process(unsigned int flag) {
  /* TO DO... */
  int rc = 0;
  struct task_struct* task = NULL;
  max_exec_task = max_mem_task = NULL;
  max_exec = max_mem = 0;

  for_each_process(task) {
    if (flag & EXEC_ERROR) {
      if (check_exec(task))
        rc |= EXEC_ERROR;
    }
    if (flag & MEM_ERROR) {
      if (check_mem(task))
        rc |= MEM_ERROR;
    }
  }

  return rc;
}


static struct timer_list det_timer;
static const unsigned int DELAY = (1.2 * HZ);
static unsigned int timeout;

static void reset_timer(int clear_flag) {
  /* reset the timer, a new delay */
  mod_timer(&det_timer, jiffies + DELAY);
  if (clear_flag)
    timeout = 0;
}

/**
 * If monitor was blocked longer than DELAY, this handler should be invoked
 * to detect and recovery failure.
 */
static void timer_handler(unsigned long data) {
  printk(KERN_EMERG "Detector: time out.\n");

  /* if continuonuly timeout, panic */
  if (++timeout >= TIMEOUT_PANIC) {
    recovery(PANIC, NULL, 0, "Monitor timeout panic");
  }
  /* check EXEC_ERROR & MEM_ERROR */
  poll_suspicious_process(EXEC_ERROR | MEM_ERROR);
  if (max_exec >= EXEC_TIMEOUT) {
    recovery(KILL, max_exec_task, 0, NULL);
  } else if (max_mem >= MAX_MEM_OCPY) {
    recovery(KILL, max_mem_task, 0, NULL);
  } else {
    recovery(NMI, NULL, 0, NULL);  /* Monitor bind to CPU0 */
  }
  reset_timer(0);
}

/**
 * Most important function in detector.c, check the system states and decide
 * whether the system should be recoverd.
 */
static void detect(void) {
  int i, cpu = 0;
  unsigned int recover_flag = 0;
  char* msg = NULL;
  /* fault1, fault5, may send NMI */
  for (i = 0; i < CPU_NR; ++i) {
    if (dm.error_code & (1 << i)) {
      /* error_code show some cpu index abnormal */
      if (dm.c[i].sys >= 95 && dm.c[i].usr < 1) {
        cpu = i;
        recover_flag = NMI;
        break;
      }
    }
  }
  if (recover_flag) {
    printk(KERN_EMERG "Hang error is detected, send NMI to CPU%d.\n", cpu);
    goto call_recover;
  }

  /* fault2, fault3, fault5, fault7.malloc, may KILL some process */
  struct task_struct* target_task = NULL;
  unsigned int poll_flag = 0, rc = 0;
  if (dm.error_code & MEM_ERROR)
    poll_flag |= MEM_ERROR;
  else
    poll_flag |= EXEC_ERROR;
  rc = poll_suspicious_process(poll_flag);
  if (rc & MEM_ERROR) {  /* fault3, fault7, should be modified!!! */
    /* fault7: iowait%, blk_proc_num, util%  is very high, use util%! */
    /* fault3: ... */
    /* fault7.malloc: memory leak... */
    if ((dm.util >= 90 || dm.blk_proc_num > MEM_BLK_PROC) && 
        max_mem >= MAX_MEM_OCPY) {
      recover_flag = KILL;
      target_task = max_mem_task;
    }
  } else if (rc & EXEC_ERROR) {
    recover_flag = KILL;
    target_task = max_exec_task;
  }
  if (recover_flag) {
    printk(KERN_EMERG "Hang error is detected, "
        "task \"%s\" (pid %i) would be killed.\n", 
        target_task->comm, target_task->pid);
    goto call_recover;
  }

  /* fault4, fault7.fork, may PANIC, to be modified!!! */
  if (dm.run_proc_num > (PER_CPU_MAX_RUN*CPU_NR) ||
      dm.blk_proc_num > MAX_BLK_PROC) {
    recover_flag = PANIC;
  }

  for (i = 0; i < CPU_NR; ++i) {  /* detect fault4 */
    if (dm.error_code & (1 << i)) {
      /* kbmemfree maybe should be replace by pswpout_per_sec! */
      if (dm.c[i].iowait > 95 && dm.kbmemfree < MEM_LOW_BOUND) {
        recover_flag = PANIC;
      }
    }
  }
  if (recover_flag) {
    msg = "Hang error is detected, would panic.";
    goto call_recover;
  }

  return;  /* no fault, return */
call_recover:
  recovery(recover_flag, target_task, cpu, msg);
}

static int recv_msg(struct sock *sk, int optval, void *user, unsigned int len) {
  if (optval == SOCKET_OPS_SET) {
    copy_from_user(&dm, user, sizeof(struct det_msg));
    switch(dm.cmd) {
      case DET_WAKEUP: {
        reset_timer(1);
        break;
      }
      case DET_ALERT: {
        printk(KERN_EMERG "Monitor alert! error_code=0x%x\n", dm.error_code);
        detect();
        break;
      }
      default:
      break;
    }
    return 0;
  } else {
    printk(KERN_EMERG "Detector: unknown optval.\n");
    return -1;
  }
}


static struct nf_sockopt_ops det_sockops = {
  .pf = PF_INET,
  .set_optmin = SOCKET_OPS_SET,
  .set_optmax = SOCKET_OPS_MAX,
  .set = recv_msg,
  .get_optmin = SOCKET_OPS_GET,
  .get_optmax = SOCKET_OPS_MAX,
  .get = NULL,
};

static int __init init_detector(void) {
  init_timer(&det_timer);
  det_timer.expires = jiffies + (4 * DELAY);  /* sar need long start time!!! */
  det_timer.data = 0;
  det_timer.function = timer_handler;
  add_timer(&det_timer);

  timeout = 0;

  int rc = nf_register_sockopt(&det_sockops);
  printk("Detector is coming! register-rc=%d\n", rc);
  return rc;
}

static void __exit exit_detector(void) {
  nf_unregister_sockopt(&det_sockops);
  del_timer_sync(&det_timer);
  printk("Detector says good bye.\n");
}

MODULE_AUTHOR("Big Tree Team");
MODULE_ALIAS("Hang detector");
MODULE_DESCRIPTION("Receive message from monitor and detect whether system hang occurs.");
MODULE_LICENSE("GPL");

module_init(init_detector);
module_exit(exit_detector);

