#include <linux/init.h>
#include <linux/module.h>
#include <linux/kernel.h>

#include <linux/sched.h>
#include <linux/thread_info.h>

#include <linux/delay.h>		// msleep()
#include <linux/notifier.h>		// struct notifier_block
#include <linux/kdebug.h>		// register_die_notifier(), die_val
#include <linux/kthread.h>		// kthread_stop()

#include <asm/atomic.h>			// atomic_read()
#include <asm/apic.h>			// apic
#include <asm/irq_vectors.h>		// NMI_VECTOR

enum recover_flag {
  PANIC = 1,
  KILL,
  NMI,
};

static int from_me = 0;

static inline int is_kthread(struct task_struct* task) {
  return task->mm == NULL;
}

static int do_panic(const char* msg) {
  panic(msg);
  return 0;
}

static int do_nmi(int cpu) {
  from_me = 1;
  apic->send_IPI_mask(cpumask_of(cpu), NMI_VECTOR);
  return 0;
}

static int do_kill(unsigned int flag, struct task_struct* task) {
  switch (task->pid) {
    case 0: {
      printk(KERN_EMERG "idle couldn't be killed!\n");
      return -1;
    }
    case 1: {
      printk(KERN_EMERG "init couldn't be killed!\n");
      return -1;
    }
    default: {
      char* ts = (task == current) ? "current" : "task";
      printk(KERN_EMERG "Recover: kill %s \"%s\" (pid %i)\n", 
          ts, task->comm, task->pid);
      if (is_kthread(task)) {
        kthread_stop(task);
      } else {
        switch (flag) {
          case NMI: {
            set_task_state(task, __TASK_STOPPED);
            schedule();
            break;
          }
          case KILL: {} /* fall through */
          default:
            task->mm->mm_users.counter = 1;
            force_sig(SIGKILL, task);
            break;
        }
      }
      return 0;
    }
  }
}

/**
 * Receive command from Detector, and execute operation.
 */
static int recovery(unsigned int flag, 
    struct task_struct * task, int cpu, const char* msg) {
  switch (flag) {
    case PANIC:
      do_panic(msg); break;
    case KILL:
      do_kill(KILL, task); break;
    case NMI:
      do_nmi(cpu); break;
    default :
      break;
  }
  return 0;
}
EXPORT_SYMBOL(recovery);

static int nmi_notifier_call(struct notifier_block* self,
    unsigned long cmd, void *__args) {
  if(from_me) {
    from_me = 0;
    do_kill(NMI, current);
  }
  return 0;
}

static struct notifier_block nmi_callback = {
  .notifier_call = nmi_notifier_call,
  .next = NULL,
  .priority = 0
};

static int __init recover_init(void) {
  printk("Recover comes\n");
  register_die_notifier(&nmi_callback);
  return 0;
}

static void __exit recover_exit(void) {
  unregister_die_notifier(&nmi_callback);
  printk("Recover exits\n");
}

MODULE_LICENSE("GPL");
MODULE_AUTHOR("S J L");
MODULE_DESCRIPTION("This is a recover!");
MODULE_ALIAS("A recover");

module_init(recover_init);
module_exit(recover_exit);

